// The TTrend component is a graphical component designed to show
// an animated plot of parameter values. As values are added old values
// scroll off the graph to the left. The trend can be viewed in a number of ways
//
// tsBar - Each value is represented as a rectangle
// tsLine - A line is drawn connecting each value
// tsScatter - A small rectangle is drawn at each value
// ts3D - Each value is represented by a 3D rectangle
// tsFilled - As tsLine except that the area below the line is filled
//
// A few options can be applied
// toValues - On tsBar and ts3D text values are added to each column.
// toTwin - Two values are displayed for each slot
//
// One area where the last option is useful is when monitoring systems with 
// demand and feedback values. Here it is important to know that a given command
// is tracked accuratley. Differences between command and feedback are especially
// well highlighted by using ts3D and toTwin
//
// Mark Dodson 3rd July 1997
//
unit Trend;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, Clipbrd;

type
  TTrendStyle = (tsBar,tsLine,tsScatter,ts3D,tsFilled);
  TTrendOption = (toValues,toTwin);
  TTrendOptions = set of TTrendOption;
  TTrend = class(TPanel)
  private
    { Private declarations }
    Fvals : array [0..99] of Integer;
    FBaseVals : array [0..99] of Integer;
    FDivisions : Integer;       
    FMinVal : Integer;
    FMaxVal : Integer;
    FStyle : TTrendStyle;
    FOptions : TTrendOptions;
    FColorBase : TColor;
    H_3d,W_3d : Integer;
    procedure SetStyle( Value:TTrendStyle );
    procedure SetMaxVal( Value:Integer );
    procedure SetMinVal( Value:Integer );
    procedure SetDivisions( Value:Integer );
    procedure SetOptions( Value:TTrendOptions );
    procedure SetColorBase( Value:TColor );
    function Brightness( OriginalColor:TColor;Change:Integer ) : TColor;
  protected             
    { Protected declarations }
    procedure Paint; Override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy;override;
    procedure Demo;
    procedure Add( NewVal:Integer );
    procedure AddBase( NewVal:Integer );
    procedure Clear;    
    procedure ClearBase;    
    function Get( i:Integer ) : Integer;
    function GetBase( i:Integer ) : Integer;
  published            
    { Published declarations }
    property MinVal : Integer read FMinVal write SetMinVal;
    property MaxVal : Integer read FMaxVal write SetMaxVal;
    property Divisions : Integer read FDivisions write SetDivisions;
    property Options : TTrendOptions read FOptions write SetOptions;
    property Style : TTrendStyle read FStyle write SetStyle;
    property ColorBase : TColor read FColorBase write SetColorBase;
  end;                                      

procedure Register;

implementation

constructor TTrend.Create(AOwner: TComponent);
var
  i : Integer;
begin
  inherited Create(AOwner);

  FMinVal := 0;
  FMaxVal := 100;
  FDivisions := 100;
  
  for i:=0 to (FDivisions - 1)  do begin
     Fvals[i] := Random( FMaxVal-FMinVal ) + FMinVal;
     FBaseVals[i] := Random( Fvals[i]-FMinVal ) + FMinVal;
  end;    

  FStyle := tsBar;
  FColorBase := clAqua;
  
  FDivisions := 20;
  H_3d := 10;
  W_3d := 10;
end;  

destructor TTrend.Destroy;
begin
  inherited Destroy;
end;  

procedure TTrend.SetStyle( Value:TTrendStyle );
begin
   if Value <> FStyle then
   begin
      FStyle := Value;
      Invalidate;
   end;
end;

procedure TTrend.SetColorBase( Value:TColor );
begin
   if Value <> FColorBase then
   begin
      FColorBase := Value;
      Invalidate;
   end;
end;

procedure TTrend.SetOptions( Value:TTrendOptions );
begin
   if Value <> FOptions then
   begin
      FOptions := Value;
      Invalidate;
   end;
end;

procedure TTrend.SetMinVal( Value:Integer );
begin
   if Value <> FMinVal then
   begin
      FMinVal := Value;
      Invalidate;
   end;
end;

procedure TTrend.SetMaxVal( Value:Integer );
begin
   if Value <> FMaxVal then
   begin
      FMaxVal := Value;
      Invalidate;
   end;
end;

procedure TTrend.SetDivisions( Value:Integer );
begin
   if Value > 100 then Value := 100;
   if Value < 2 then Value := 2;
   if Value <> FDivisions then
   begin
      FDivisions := Value;
      Invalidate;
   end;
end;

function TTrend.Brightness( OriginalColor:TColor;Change:Integer ) : TColor;
var
   R,G,B : Integer; // Colour components
   nR,nG,nB : Integer; // New Colour components
begin
   R := OriginalColor and $ff;
   G := (OriginalColor and $ff00) shr 8;
   B := (OriginalColor and $ff0000) shr 16;
   nR := R + Change;
   nG := G + Change;
   nB := B + Change;
   // Limit Brightness change to prvent colour change
   if nR > $ff then nR := $ff;
   if nG > $ff then nG := $ff;
   if nB > $ff then nB := $ff;
   if nR < 0 then nR := 0;
   if nG < 0 then nG := 0;
   if nB < 0 then nB := 0;
   Brightness := (nB shl 16) + (nG shl 8) + nR;
end;

procedure TTrend.Demo;
var
   i : Integer;
begin
   // Shift old data left
   for i:=0 to (FDivisions-2) do begin
      Fvals[i] := Fvals[i+1];
      FBaseVals[i] := FBaseVals[i+1];
   end;

   // Create new data item
   Fvals[FDivisions - 1] := Random( FMaxVal-FMinVal ) + FMinVal;
   FBaseVals[FDivisions - 1] := Random( FMaxVal-FMinVal ) + FMinVal;
                              
   // Redraw
   Invalidate;
end;

procedure TTrend.Add( NewVal:Integer );
var
   i : Integer;
begin
   // Shift old data left
   for i:=0 to (FDivisions-2) do begin
      Fvals[i] := Fvals[i+1];
   end;

   // Add new data item
   if( NewVal > FMaxVal ) then NewVal := FMaxVal
   else if( NewVal < FMinVal ) then NewVal := FMinVal;
   
   Fvals[FDivisions - 1] := NewVal;

   // Redraw
   Invalidate;
end;

procedure TTrend.AddBase( NewVal:Integer );
var
   i : Integer;
begin
   // Shift old data left
   for i:=0 to (FDivisions-2) do begin
      FBaseVals[i] := FBaseVals[i+1];
   end;

   // Add new data item
   if( NewVal > FMaxVal ) then NewVal := FMaxVal
   else if( NewVal < FMinVal ) then NewVal := FMinVal;
   
   FBaseVals[FDivisions - 1] := NewVal;

   // Redraw
   Invalidate;
end;

procedure TTrend.Clear;
var
   i : Integer;
begin
   // Shift old data left
   for i:=0 to (FDivisions-1) do begin
      Fvals[i] := 0;
   end;

   // Redraw
   Invalidate;
end;

procedure TTrend.ClearBase;
var
   i : Integer;
begin
   // Shift old data left
   for i:=0 to (FDivisions-1) do begin
      FBaseVals[i] := 0;
   end;

   // Redraw
   Invalidate;
end;

function TTrend.Get( i:Integer ) : Integer;
begin
   if ( (i >= 0) and (i < 100) ) then
      Get := Fvals[i]
   else
      Get := 0;
end;

function TTrend.GetBase( i:Integer ) : Integer;
begin
   if ( (i >= 0) and (i < 100) ) then
      GetBase := FBaseVals[i]
   else
      GetBase := 0;
end;

procedure TTrend.Paint;
var
   BarWidth : Variant;
   BorderWidth : Integer;
   i,x,y,w,y1,y2,LastX : Integer;
   TheImage : TBitmap;
   BrushColor : TColor;
begin
   TheImage := TBitmap.Create;
   try
   begin
      TheImage.Height := Height;
      TheImage.Width := Width;
   
      with TheImage.Canvas do begin
         Brush.color := clWhite;
         Rectangle( 0,0,Width,Height ); // Draw background
         BorderWidth := 3;
         BarWidth := (Width - 2*BorderWidth) / FDivisions;
         Brush.Assign( self.Brush );
         case FStyle of
            tsScatter:
            begin
               Pen.Color := Brush.Color;
               for i:=0 to (FDivisions - 1) do
               begin
                  Pen.Color := Color;
                  Brush.Color := Color;
                  w := BarWidth / 4;
                  if w < 1 then w := 1;
                  x := BorderWidth + (i*BarWidth) + (2 * w);
                  y := ((Fvals[i]-FMinVal)*(Height - 2*BorderWidth - (2 * w)) div (FMaxVal-FMinVal)) 
                       + BorderWidth + w;
                  Rectangle( x - w,
                             Height - (y - w),
                             x + w,
                             Height - (y + w) );

                  if( toTwin in FOptions ) then begin
                     Pen.Color := ColorBase;
                     Brush.Color := ColorBase;
                     y := ((FBaseVals[i]-FMinVal)*(Height - 2*BorderWidth - (2 * w)) div (FMaxVal-FMinVal)) 
                          + BorderWidth + w;
                     Rectangle( x - w,
                                Height - (y - w),
                                x + w,
                                Height - (y + w) );
                  end;
               end;
            end;
            
            tsLine:
            begin
               Pen.Color := Color;
               moveto( BorderWidth + (BarWidth div 2),
                       Height - (((Fvals[0]-FMinVal)*(Height - 2*BorderWidth) div (FMaxVal-FMinVal)) 
                                      + BorderWidth) );
               for i:=1 to (FDivisions - 1) do
               begin
                  lineto( BorderWidth + (i*BarWidth) + (BarWidth div 2),
                          Height - (((Fvals[i]-FMinVal)*(Height - 2*BorderWidth) div (FMaxVal-FMinVal)) 
                                      + BorderWidth) );
               end;
               
               if( toTwin in FOptions ) then begin
                  Pen.Color := ColorBase;
                  Brush.Color := ColorBase;
                  moveto( BorderWidth + (BarWidth div 2),
                          Height - (((FBaseVals[0]-FMinVal)*(Height - 2*BorderWidth) div (FMaxVal-FMinVal)) 
                                      + BorderWidth) );
                  for i:=1 to (FDivisions - 1) do
                  begin
                     lineto( BorderWidth + (i*BarWidth) + (BarWidth div 2),
                             Height - (((FBaseVals[i]-FMinVal)*(Height - 2*BorderWidth) div (FMaxVal-FMinVal)) 
                                         + BorderWidth) );
                  end;
               end;
            end;

            tsFilled:
            begin
               Pen.Color := Brush.Color;
               for i:=0 to (FDivisions - 2) do
               begin
                  x := BorderWidth + (i*BarWidth) + (BarWidth / 2);
                  y1 := ((Fvals[i]-FMinVal)*(Height - 2*BorderWidth) div (FMaxVal-FMinVal)) + BorderWidth;
                  y2 := ((Fvals[i+1]-FMinVal)*(Height - 2*BorderWidth) div (FMaxVal-FMinVal)) + BorderWidth;
                  Polygon( [Point( x,Height - BorderWidth ),
                           Point( x + BarWidth, Height - BorderWidth ),
                           Point( x + BarWidth, Height - y2 ),
                           Point( x, Height - y1 )] );
               end; 
            end;
            
            ts3D:
            begin
               BarWidth := (Width - 2*BorderWidth - W_3d) / FDivisions;
               BrushColor := Brush.Color;
               Font.Height := 8;
               Font.Name := 'Arial';
               x := BorderWidth;
               for i:=0 to (FDivisions - 1) do
               begin
                  LastX := x;
                  x := BorderWidth + ((i+1)*BarWidth);
                  y := ((Fvals[i]-FMinVal)*(Height - 2*BorderWidth - H_3d) div (FMaxVal-FMinVal)) 
                       + BorderWidth;
                  if( toTwin in FOptions ) then
                     y1 := ((FBaseVals[i]-FMinVal)*(Height - 2*BorderWidth - H_3d) div (FMaxVal-FMinVal)) 
                             + BorderWidth
                  else
                     y1 := BorderWidth;

                  // Make sure y is the smallest
                  if y < y1 then begin
                     y2 := y;
                     y := y1;
                     y1 := y2;
                  end;
                  Brush.Color := BrushColor;
                  Pen.Color := Brush.Color;
                  Polygon( [Point( LastX, Height - y1 ),
                           Point( x, Height - y1 ),
                           Point( x, Height - y ),
                           Point( LastX, Height - y )] );
                  // Draw side perspective extension in darker colour
                  Brush.Color := Brightness( BrushColor, -60 );
                  Pen.Color := Brush.Color;
                  Polygon( [Point( x, Height - y1 ),
                           Point( x + W_3d, Height - (y1 + H_3d) ),
                           Point( x + W_3d, Height - (y + H_3d) ),
                           Point( x, Height - y )] );
                  // Draw top perspective extension in lighter colour
                  Brush.Color := Brightness( BrushColor, 60 );
                  Pen.Color := Brush.Color;
                  Polygon( [Point( x, Height - y ),
                           Point( x + W_3d, Height - (y + H_3d) ),
                           Point( LastX + W_3d, Height - (y + H_3d) ),
                           Point( LastX, Height - y )] );
                  if( toValues in FOptions ) then begin
                     Brush.Color := clWhite;
                     if( toTwin in FOptions ) then
                        TextOut( LastX + (BarWidth div 2) - (TextWidth( IntToStr( abs(Fvals[i]-FBaseVals[i]) ) ) div 2),                           Height - y - (Font.Height),
                                 IntToStr( abs( Fvals[i] - FBaseVals[i]) ) )
                     else
                        TextOut( LastX + (BarWidth div 2) - (TextWidth( IntToStr( Fvals[i] ) ) div 2),                           Height - y - (Font.Height),
                                 IntToStr( Fvals[i] ) );
                  end;
               end;
            end;
            
            tsBar:
            begin
               BrushColor := Brush.Color;
               Font.Name := 'Arial';
               Font.Height := 8;
               for i:=0 to (FDivisions - 1) do
               begin
                  x := BorderWidth + (i*BarWidth);
                  if ( toValues in FOptions ) then
                     y := ((Fvals[i]-FMinVal)*(Height - 2*BorderWidth - Font.Height) div (FMaxVal-FMinVal)) 
                          + BorderWidth
                  else
                     y := ((Fvals[i]-FMinVal)*(Height - 2*BorderWidth) div (FMaxVal-FMinVal)) + BorderWidth;
                  Brush.Color := BrushColor;
                  rectangle( x,
                             Height - BorderWidth,
                             x + BarWidth,
                             Height - y );
                  if( toValues in FOptions ) then begin
                     Brush.Color := clWhite;
                     TextOut( x + (BarWidth div 2) - (TextWidth( IntToStr( Fvals[i] ) ) div 2),
                              Height - y - (Font.Height),
                              IntToStr( Fvals[i] ) );
                  end;
               end;
            end; // of case FStyle
         end;
      end;

      // Now copy onto screen canvas
      Canvas.CopyMode := cmSrcCopy; 
      Canvas.CopyRect( ClientRect,TheImage.Canvas,ClientRect );
   end;
   finally
      TheImage.Destroy;
   end;
end;

procedure Register;
begin
  RegisterComponents('Mark', [TTrend]);
end;

end.
 